=begin
 Copyright (c) 2006, Artur Merke <artur.merke@udo.edu> 

 This is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 It is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
=end

require 'yaml'

#version 1.4

def player_name(p)
  if p.dtkv
    name= p.dtkv.vname + " " + p.dtkv.nname
  else
    name= String.new(p.key)
    name.sub(/(\w)([^_]){3,}_/,'\1_')
    name.sub!(/_/,". ")
    name.gsub!(/_/," ")
  end
  name
  #name + " (#{p.white_black_balance})" #test
end

def referee_name(p)
  player_name(p) + " (#{p.stat.num_games_as_referee})"
  #player_name(p) + "(#{p.real_rank})" #test
end

##############################################################################
##############################################################################
##############################################################################

class Options
  def can_be_referee(key)
    return true if not @no_referee
    return (not @no_referee.include?(key))
  end

  def use_buchholz
    return @use_buchholz if @use_buchholz
    #default value
    return true
  end

  def input_dir
    return @input_dir if @input_dir
    #default value
    return './in/'
  end

  def output_dir
    return @output_dir if @output_dir
    #default value
    return './out/'
  end
  
  def random_results
    return @random_results if @random_results
    #default value
    return false
  end

  def fontsize_ranking
    return @fontsize_ranking if @fontsize_ranking
    #default value
    return 9
  end

  def fontsize_schedule
    return @fontsize_schedule if @fontsize_schedule
    #default value
    return 12
  end

  def fontsize_crosstable
    return @fontsize_crosstable if @fontsize_crosstable
    #default value
    return 12
  end

end

##############################################################################


class HTML

  attr_accessor :fontsize

  def initialize
    @fontsize= 12
  end

  def print_document_begin(out, title=nil)
    #xhtml 1.1 header
    out.print '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"' 
    out.print "\n",'   "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">'
    out.print "\n<html>"
    out.print "\n<head>"
    out.print "\n",'<meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1"/>'
    if true
      out.print "\n",'<style type="text/css">'
      out.print "\n",'/*<![CDATA[*/'
      out.print "\n<!--"
out.print <<EOF 
body {
color: black !important; 
font-family: "Arial", sans-serif !important; 
font-size: #{fontsize}pt !important;
}
table {
color:#000000; 
background-color:#ffffff; 
border-style:solid;
border-collapse:collapse;
#border-spacing:0px;
#padding:9px;
}
td {
border-style:solid;
border-width:1px;
}
tr {
#border-style:solid;
#border-width:1px;
}
EOF
      out.print "\n-->"
      out.print "\n",'/*]]>*/'
      out.print "\n</style>"
    end  
    out.print "\n<title>"
    out.print title if title
    out.print "</title>"
    out.print "\n</head>"
    out.print "\n<body>"
  end

  def print_document_end(out)
    #out.print "\n<p>",$time_stamp,"</p>"
    out.print "\n</body>"
    out.print "\n</html>"
  end
end

##############################################################################

class Player
  attr_reader :key
  #rank is always unique
  attr_accessor :rank
  #two players have the same real_rank if they rankings are equal
  attr_accessor :real_rank
  attr_accessor :stat
  attr_reader :dummy, :can_be_ref
  attr_accessor :tmp  #for storing temporary data (like past ranking)
  attr_accessor :dtkv #data from DTKVranking::PlayerInfo
  attr_accessor :white_black_balance #count for games player with white/black

  def initialize( k )
    @key= k
    @rank= 0
    @real_rank= -1
    @stat= nil
    @dummy= false
    @can_be_ref= true
    @white_black_balance= 0
    @tmp=nil
    @dtkv= nil

    #these players will never be referees:
    @can_be_ref= $options.can_be_referee(k)

    if k == 'dummy'
      @dummy= true 
      @can_be_ref= false
    end
  end
end

##############################################################################

class DTKVranking
  class PlayerInfo
    attr_reader :rank, :key, :vname, :nname, :club
    def initialize(r,k,v,n,c)
      @rank= r
      @key= k
      @vname= v
      @nname= n
      @club= c
    end
  end

  def initialize(fname)
    @arr=[]
    read(fname)
  end

  def read(fname)
    comment = /^\s*#/
    blank_line = /^\s*$/
    #data_line=/^(\d+),([^,]+),([^,]*),([^,]+),([^,]*)$/
    data_line=/^(\d+),(.+?),(.+?),(.+?),(.*?)$/
    IO.foreach(fname) { |line|
      next if line =~ comment
      next if line =~ blank_line
      if not r=data_line.match(line)
        raise "\n\nwrong format line (#{fname}):\n"+ line
        next
      end

      @arr << PlayerInfo.new(r[1].to_i,r[2],r[3],r[4],r[5])
    }
  end

  def get_or_nil(key)
    @arr.each{ |p|
      return p if p.key == key
    }
    return nil
  end
end

##############################################################################

class PlayersDB
  def initialize(fname,dtkv_ranking)
    @dummy=nil
    @players={}
    @dtkv_ranking= dtkv_ranking
    read(fname)
    if @players.size % 2 == 1
      print "\nadding a dummy player, so that the number of players is an even number"
      k= 'dummy'
      @dummy= Player.new(k)
      @players[k]= @dummy
    end
  end

  def put( k )
    pdata= @players.fetch(k, nil)
    raise "player #{k} already in" if pdata
    pdata= Player.new(k)
    pdata.dtkv= @dtkv_ranking.get_or_nil(k)
    if not pdata.dtkv
      print "\n\nWarning: player ", k , " not in the dtkv ranking"
    end
    @players[k]= pdata
  end

  def get_or_nil( k )
    @players.fetch(k, nil)
  end
    
  def get( k )
    pdata= @players.fetch(k, nil)
    raise "no such player \"#{k}\"" unless pdata
    return pdata
  end

  def show
    print "\nPlayersDB"
    @players.each{ |k,p|
      print "\n", k , " -> ", p.key , " rank= ", p.rank
    }
  end

  def each
    @players.each{ |k,v|
      yield v
    }
  end
  
  private

  def read( fname )
    arr= players= YAML::load( File.open(fname) )
    arr.each_with_index {|p,i|
      #break if i > 7 #debug
      #break if i > 8 #debug
      #break if i > 20 #debug
      put(p)
    }
  end
end

##############################################################################

class Game
  attr_reader :p1, :p2, :ref
  attr_accessor :res1, :res2

  def initialize(p1, p2, res1, res2, ref=nil)
    @p1= p1
    @p2= p2
    @res1= res1
    @res2= res2
    @ref= ref
  end

  def set_referee( ref )
    @ref= ref
  end
end

class GamesDB
  attr_accessor :round

  def initialize(players_db= nil)
    @games=[]
    @players_db= players_db
    @round= 0
  end
  
  def put( g )
    @games << g
  end

  def each 
    @games.each{ |e|
      next if Marker === e
      yield e
    }
  end

  def debug
    print "\nOpponents of P_Funke"
    sum1=0
    sum2=0
    pp= nil
    each{ |g|
      p= nil
      p= g.p2 if g.p1.key == 'P_Funke' 
      p= g.p1 if g.p2.key == 'P_Funke' 
      pp= g.p1 if g.p1.key == 'P_Funke' 
      pp= g.p2 if g.p2.key == 'P_Funke' 
      
      next unless p
      print "\ngame ", g.p1.key, " ", g.p2.key, " ", g.res1, " ", g.res2
      print "\n", p.key, "avg points", p.stat.get_avg_points, " W:", p.stat.num_wins, " G: ", p.stat.num_games
      sum1+= p.stat.num_wins*2
      sum2+= p.stat.num_games
    }
    print "\n sum1= ", sum1, " sum2= ", sum2 , " frac= ", sum1.to_f/sum2
    print "\n frac2= %0.03f" % (sum1.to_f/sum2.to_f)
    print "\n stat:"
    print "\n pp.stat num_opp_points: ", pp.stat.num_opp_points
    print "\n pp.stat num_opp_games: ", pp.stat.num_opp_games
    print "\n pp.stat frac= ", pp.stat.num_opp_points.to_f/pp.stat.num_opp_games
  end
  
  def each_with_round 
    r= 0
    @games.each{ |e|
      if Marker === e
        r= e.round
        next
      end
      yield e,r
    }
  end
  
  def concat( gdb )
    @games.concat( gdb.games )
  end

  def show
    print "\nGamesDB"
    each{ |g|
      print "\n", g.p1.key, " ", g.p2.key, " ", g.res1, " ", g.res2
    }
  end
  
  NL="\n"
  LEV=[NL,"",""]

  def show_game_against_dummy(out, p, idx)
    #show the player without a game also in the table!
    out.print LEV[0],'<tr align="center"'
    out.print ' bgcolor="#eeeeee"' if ( idx % 2 == 0 )
    out.print '>'
    out.print LEV[1],'<td/>'
    out.print LEV[1],'<td>', player_name(p),'</td>' 
    out.print LEV[1],'<td/>'
    out.print LEV[1],'<td>spielfrei</td>'
    out.print LEV[1],'<td/>'
    out.print LEV[1],'</tr>'
  end

  def show_html(out, mode= :last)
    col= nil
    html= HTML.new
    html.fontsize= $options.fontsize_schedule
    html.print_document_begin(out)
   
    idx= 0
    table_idx= 0
    game_against_dummy= nil

    open= false
    if mode == :last
      r= @round
    else
      r= -1
    end
    schedule_for_round(r) { |g|
      if Marker === g
        if open
          if game_against_dummy
            show_game_against_dummy(out, game_against_dummy, idx+1)
          end
          
          out.print LEV[0],'</tbody>'
          out.print LEV[0],'</table>'
        end
        open= true
        idx= 0
        table_idx= 0
        game_against_dummy= nil
        
        if g.pass === 1
          out.print LEV[0],"<h1> Runde: #{g.round}</h1>"
        end

        out.print LEV[0],"<h2> Durchgang: #{g.pass}</h2>"
        out.print LEV[0],'<table>'
        out.print LEV[0],'<tbody>'

        out.print LEV[0],'<tr align="center">'
        out.print LEV[1],'<td> Platte </td>'
        out.print LEV[1],'<td> Spieler</td>'
        out.print LEV[1],'<td> Ergebnis </td>'
        out.print LEV[1],'<td> Spieler </td>'
        out.print LEV[1],'<td> Schiedsrichter </td>'
        out.print LEV[0],'</tr>'
        next
      end

      #next if g.p1.dummy or g.p2.dummy
      if g.p1.dummy or g.p2.dummy
        game_against_dummy= g.p1
        game_against_dummy= g.p2 if g.p1.dummy
        next
      end

      out.print LEV[0],'<tr align="center"'
      idx+= 1
      out.print ' bgcolor="#eeeeee"' if ( idx % 2 == 0 )
      out.print '>'

      table_idx+= 1

      out.print LEV[1],'<td>', table_idx  ,'</td>'
      out.print LEV[1],'<td>', player_name(g.p1),'</td>'
      out.print LEV[1],'<td>'
      out.print g.res1,":", g.res2 if g.res1 and g.res2
      out.print LEV[1],'</td>'
      out.print LEV[1],'<td>',  player_name(g.p2),'</td>'
      if  g.ref 
        out.print LEV[1],'<td>',  referee_name(g.ref)
        out.print '</td>' 
      else
        out.print LEV[1],'<td/>'
      end
      out.print LEV[0],'</tr>'
    }

    if game_against_dummy
      show_game_against_dummy(out, game_against_dummy, idx+1)
    end

    if open
      out.print LEV[0],'</tbody>'
      out.print LEV[0],'</table>'
    end
    html.print_document_end(out)
  end

  def size
    @games.size
  end

  def read( fname )
    raise "\nno players db provided!" unless @players_db
    comment = /^\s*#/
    blank_line = /^\s*$/
    #name= /[a-zA-Z_]+/
    #game_line=/^\s*(\w+)\s+(\w+)\s+(\d+)\s+(\d+)\s*(\w*)\s*$/
    #game_line=/^\s*([a-zA-Z_]+)\s+([a-zA-Z_]+)\s+(\d+)\s+(\d+)\s*([a-zA-Z_]*)\s*$/

    game_line=/^\s*(.+?)\s+(.+?)\s+(\d+)\s+(\d+)\s*(.*?)\s*$/

    iter_line=/^\s*\\round (\d+)\s+\(\s*([12]){0,1}\s*\)\s*$/
    
    IO.foreach(fname) { |line|
      next if line =~ comment
      next if line =~ blank_line

      if r= iter_line.match(line)
        @games << Marker.new(r[1].to_i, r[2].to_i)
        @round= r[1].to_i
        next
      end
      
      if r= iter_line.match(line)
        @round= r[1].to_i
        next
      end

      if not r=game_line.match(line)
        raise "\n\nskipping wrong format line (#{fname}):\n"+ line
        next
      end

      p1, p2, res1, res2, ref = r[1,5]
      ref= nil if ref.size == 0
      #print "\ndebug: ", p1," ",p2," ", res1, " ", res2, " [", ref , "]"
      res1= res1.to_i
      res2= res2.to_i

      ref= @players_db.get(ref) if ref
      p1= @players_db.get(p1)
      p2= @players_db.get(p2)
      @games << Game.new( p1, p2, res1, res2, ref)
      p1.white_black_balance += 1 unless p2.dummy
      p2.white_black_balance -= 1 unless p1.dummy
    }
    #@name_to_fullname.each{ |k,d|  print "\nkey=",k," data=",d }
    #exit
  end

  def show_text(out)
    max_len= 0
    dummygame= nil
    @games.each{ |g|
      next if Marker === g
      l= g.p1.key.length
      max_len= l if l > max_len
      l= g.p2.key.length
      max_len= l if l > max_len
    }
    
    @games.each{ |g|
      if Marker === g
        if g.pass == 1
          out.print "\n###################################################"
        end  
        out.print "\n\n\\round #{g.round} (#{g.pass})\n"
        next
      end

      out.print "\n"
      #out.print "\n", g.p1.key, " ", g.p2.key, " ", g.res1, " ", g.res2 
      out.print g.p1.key
      out.print " "*(max_len -g.p1.key.length)
      out.print "  "
      out.print g.p2.key
      out.print " "*(max_len -g.p2.key.length)
      if g.res1 and g.res2
        out.print "  ", g.res1, " ", g.res2 
      else
        if g.p1.dummy or g.p2.dummy
          out.print "  0 0    "
        else
          out.print "         "
        end
      end
      out.print "    ", g.ref.key if g.ref
    }
  end

  # def append(fname)
  #   #out= File.open(fname,"w")
  #   out= File.open(fname,File::CREAT|File::APPEND|File::WRONLY)
  #   show_text(out)
  #   out.close
  # end
  
  def write(fname)
    out= File.open(fname,"w")
    #out= File.open(fname,File::CREAT|File::APPEND|File::WRONLY)
    show_text(out)
    out.close
  end

  def generate_random_results
    each{ |g|
      if false
        res1= (rand * 5.0).round.to_i #(rand * 6.0 + res1*3.0).to_i
        res2= (rand * 5.0).round.to_i #(rand * 6.0 + res2*3.0).to_i
      else
        wp= 0.5
        if g.p1.dtkv and g.p2.dtkv
          wp= 0.6 if g.p1.dtkv.rank < g.p2.dtkv.rank
          wp= 0.7 if g.p1.dtkv.rank + 10 < g.p2.dtkv.rank
          wp= 0.8 if g.p1.dtkv.rank + 20 < g.p2.dtkv.rank
          wp= 0.9 if g.p1.dtkv.rank + 40 < g.p2.dtkv.rank
          wp= 0.4 if g.p2.dtkv.rank < g.p1.dtkv.rank
          wp= 0.3 if g.p2.dtkv.rank + 10 < g.p1.dtkv.rank
          wp= 0.2 if g.p2.dtkv.rank + 20 < g.p1.dtkv.rank
          wp= 0.1 if g.p2.dtkv.rank + 40 < g.p1.dtkv.rank
        elsif g.p1.dtkv
          wp= 0.95
        else
          wp= 0.05
        end
        if rand <= wp
          res1= 1
          res2= 0
        else
          res1= 0
          res2= 1
        end
      end
      g.res1= res1
      g.res2= res2
    }
  end
  
  def shift_best_referee(p1, p2, referees)
    idx= 0
    vl= 'vereinslos'
    referees.each_with_index{|r,i|
      #print "\n   idx= ", idx
      if p1.dtkv and p2.dtkv and (p1.dtkv.club == p2.dtkv.club)
        idx= i
        break
      end
      if p1.dtkv and r.dtkv and (p1.dtkv.club == r.dtkv.club)
        next if p1.dtkv.club != vl
      end
      if p2.dtkv and r.dtkv and (p2.dtkv.club == r.dtkv.club)
        next if p2.dtkv.club != vl
      end
      idx= i
      break
    }
    referees.delete_at(idx)
  end

  def make_schedule
    #computing referees
    odd_refs=[]
    even_refs=[]
    odd_games= 0
    even_games= 0
    arr= @games
    @games=[]
    
    arr.each_with_index{ |g,i|
      if i % 2 == 1
        odd_refs << g.p1 if g.p1.can_be_ref
        odd_refs << g.p2 if g.p2.can_be_ref
        odd_games+= 1 if not (g.p1.dummy or g.p2.dummy)
      else
        even_refs << g.p1 if g.p1.can_be_ref
        even_refs << g.p2 if g.p2.can_be_ref
        even_games+= 1 if not (g.p1.dummy or g.p2.dummy)
      end
    }

    odd_refs.sort!{ |p1,p2| 
      p1.stat.num_games_as_referee <=> p2.stat.num_games_as_referee 
    }
    odd_refs= odd_refs.slice(0, even_games)
    odd_refs.sort!{ |p1,p2|
      p1.rank <=> p2.rank
    }

    even_refs.sort!{ |p1,p2|
      ret= p1.stat.num_games_as_referee <=> p2.stat.num_games_as_referee 
    }
    even_refs= even_refs.slice(0, odd_games)
    even_refs.sort!{ |p1,p2|
      p1.rank <=> p2.rank
    }
    
    # print "\n########################"
    # print "\n########################"
    # print "\n########################"
    # print "\neven_ref (", even_refs.size, ")"

    arr.each_with_index{|g,i|
      next if g.p1.dummy or g.p2.dummy
      if i % 2 == 0
        #g.set_referee( odd_refs.shift )
        g.set_referee( shift_best_referee( g.p1, g.p2, odd_refs ) )
      else
        #g.set_referee( even_refs.shift )

        # print "\ngame= ", i, "\n"
        # print "\nplayer= ", g.p1.key, ", ", g.p2.key
        # print "  p1.c  (", g.p1.dtkv.club ,")"if g.p1.dtkv
        # print "  p2.c  (", g.p2.dtkv.club, ")" if g.p2.dtkv
        # print "\nref= "
        # even_refs.each{|r| 
        #   print " ", r.key
        #   print "(",r.dtkv.club,")" if r.dtkv 
        # }

        g.set_referee( shift_best_referee( g.p1, g.p2, even_refs ) )
      end
    }

    #now set the schedule

    @games << Marker.new(@round,1)
    arr.each_with_index{ |g,i|
      if i % 2 == 0
        @games << g
      end
    }
    @games << Marker.new(@round,2)
    arr.each_with_index{ |g,i|
      if i % 2 == 1
        @games << g
      end
    }
  end

protected

  attr_reader :games

  def schedule_for_round( rrr )
    r = -1
    @games.each{ |e|
       if Marker === e
         r= e.round
       end
       if rrr > 0
         next if r != rrr
       end  
       yield e
    }
  end

  class Marker
    attr_reader :round, :pass
    def initialize(r, p=nil)
      @round= r
      @pass= p
    end
  end
end

##############################################################################

class Ranking
  class PlayerGamesStat 
    attr_reader :num_games, :num_wins, :num_defeats, :num_draws, 
                :num_goals, :num_goals_against,
                :num_games_as_referee,
                :num_opp_games, :num_opp_points

    def initialize
      @num_games= 0;
      @num_wins= 0;
      @num_defeats= 0;
      @num_draws= 0;
      @num_goals= 0;
      @num_goals_against= 0;
      @num_games_as_referee= 0;

      #needed for buchholz value
      @num_opp_points= 0;
      @num_opp_games= 0;
    end

    def add_game(res1,res2)
      @num_games += 1;
      @num_wins += 1 if res1 > res2;
      @num_draws += 1 if res1 == res2;
      @num_defeats += 1 if res1 < res2;
      @num_goals += res1;
      @num_goals_against += res2;
    end
    
    def add_game_as_referee
      @num_games_as_referee+= 1
    end

    def add_opp_game( stat )
      @num_opp_games+= stat.num_games
      @num_opp_points+= stat.num_wins*2 + stat.num_draws
    end

    def reset_opp_games
      @num_opp_games= 0
      @num_opp_points= 0
    end

    def get_avg_points
      return get_avg(get_num_points,@num_games)
    end

    def get_avg_goals
      return get_avg(@num_goals,@num_games)
    end

    def get_avg_goals_against
      return get_avg(@num_goals_against,@num_games)
    end
    
    def get_buchholz_value
      return get_avg_01(@num_opp_points,@num_opp_games)
    end

    def <=>( pstat)
      def avg(n1,d1, n2, d2)
        return n1 * d2 <=> n2 * d1 if d1 != 0 and d2 != 0
        n1 <=> n2
      end
      def avg_01(n1,d1, n2, d2)
        #the fractions are rounded to their first position after decimal point 
        if d1 != 0 and d2 != 0
          f1= (n1.to_f*10.0/d1.to_f).to_i
          f2= (n2.to_f*10.0/d2.to_f).to_i
          return f1 <=> f2
        end
        n1 <=> n2
      end

      res= avg(get_num_points, @num_games, pstat.get_num_points, pstat.num_games)
      return res if res != 0
      if $options.use_buchholz
        res= avg(@num_opp_points, @num_opp_games, pstat.num_opp_points, pstat.num_opp_games)
        #res= avg_01(@num_opp_points, @num_opp_games, pstat.num_opp_points, pstat.num_opp_games)
        return res if res != 0
      end
      res = avg(get_goal_diff, @num_games, pstat.get_goal_diff, pstat.num_games)
      return res if res != 0
      res = avg(@num_goals, @num_games, pstat.num_goals, pstat.num_games)
      return res
    end

    def get_num_points()
      return @num_wins*2 + @num_draws
    end

    def get_num_points_against()
      return @num_defeats*2 + @num_draws
    end
    
    def get_avg_goal_diff
      return get_avg(@num_goals-@num_goals_against,@num_games)
    end

    protected

    def get_goal_diff
      return num_goals - num_goals_against
    end

    def get_point_diff
      return get_num_points - get_num_points_against
    end

    def get_avg(n,d)
      return '0.00' if d == 0
      "%.03f" % (n.to_f / d) 
    end
    
    def get_avg_01(n,d)
      if true #debug
        return '0.00' if d == 0
        return "%.03f" % (n.to_f / d) 
      end  
      return '0.0' if d == 0
      #r= ( ( (n.to_f *10.0)/d.to_f ).to_i ).to_f / 10.0
      r=  ( (n.to_f * 10.0) / d.to_f ).to_i
      
      return (r / 10).to_s + "." + (r % 10).to_s
    end
  end

  def initialize(players_db)
    @players_db= players_db
    @arr=[]
    @players_db.each{ |p|
      p.rank= @arr.size
      p.stat= PlayerGamesStat.new
      @arr << p
    }
  end
  
  def compute_ranking_from_games(games_db)
    return if games_db.size <= 0

    @arr.each{ | p |
      p.stat= PlayerGamesStat.new
      p.tmp=[]
    }

    cur_round= 1
    games_db.each_with_round{|g,r|
      next if g.p1.dummy
      next if g.p2.dummy

      if cur_round < r 
        cur_round= r

        #compute buchholz values
        @arr.each{ | p | p.stat.reset_opp_games }
        games_db.each_with_round{|g_tmp,r_tmp|
          break if r_tmp >= r
          g_tmp.p1.stat.add_opp_game( g_tmp.p2.stat )
          g_tmp.p2.stat.add_opp_game( g_tmp.p1.stat )
        }

        sort_ranking_using_stat
        @arr.each{ | p |
          p.tmp << p.real_rank
        }
      end

      g.p1.stat.add_game( g.res1, g.res2 )
      g.p2.stat.add_game( g.res2, g.res1 )
      if g.ref 
        g.ref.stat.add_game_as_referee
      end
    }
    
    #buchholz
    @arr.each{ | p | p.stat.reset_opp_games }
    games_db.each{|g|
      next if g.p1.dummy
      next if g.p2.dummy
      g.p1.stat.add_opp_game( g.p2.stat )
      g.p2.stat.add_opp_game( g.p1.stat )
    }

    sort_ranking_using_stat

    @arr.each{ | p |
      p.tmp << p.real_rank
    }
  end
  
  def compute_ranking_from_dtkv_ranking
    max_rank= 0
    @arr.each{ | p |
      max_rank= p.dtkv.rank if p.dtkv and p.dtkv.rank > max_rank
    }
    @arr.each{ | p |
      if p.dtkv
        p.rank= p.dtkv.rank 
      else
        p.rank= max_rank+1
      end
      p.rank= max_rank+2 if p.dummy
    }
    @arr.sort!{ |p1,p2| (p1.rank <=> p2.rank) }
    @arr.each_with_index{ | p , i |
      p.rank= i
      p.real_rank= i+1
    }
  end

  def interlace_ranking
    #change [0,1,2,3] to
    #       [2,1,0,3]
    step= @arr.size/2
    step= (@arr.size-2)/2 if @arr.last.dummy
    #print "\n step= ",step
    0.step(step-1,2) { |i|
      @arr[i+1], @arr[i+step] = @arr[step+i], @arr[i+1]
    }
  end

  def size
    @arr.size
  end

  def [](idx)
    raise "\nrange error" if idx >= @arr.size or idx < 0
    @arr[idx]
  end
  
  def show
    print "\nRanking"
    @arr.each{ |p|
      print "\n%2d" % p.real_rank
      print "%15s" % p.key
      print " ", p.stat.get_avg_points
      print " ", p.stat.get_buchholz_value if $options.use_buchholz
      print " ", p.stat.get_avg_goal_diff
      print " " if p.stat.get_avg_goal_diff[0] != "-"[0]
      print " ", p.stat.get_avg_goals
      print " " if p.stat.get_avg_goals[0] != "-"[0]
=begin
      if p.tmp
        print " |"
        p.tmp.each_with_index{ |old_rank, idx|
          #print " ",idx+1,":", old_rank
          print " ", "%2d" % old_rank
        }
      end
=end
    }
  end

  def show_yaml(out)
    @arr.each{ |p|
      out.print "\n  - ",p.key
    }
  end

  NL="\n"
  LEV=[NL,"",""]

  def show_html ( out, num=nil )
    col= nil
    html= HTML.new
    html.fontsize= $options.fontsize_ranking
    html.print_document_begin(out)
    if num and num > 0
      out.print LEV[0], "<h1> Rangliste nach #{num}. Runde"
      out.print "n" if num > 1
      out.print "</h1>"
    else
      out.print LEV[0], "<h1> Rangliste </h1>"
    end

    out.print LEV[0],'<table>'
    out.print LEV[0],'<tbody>'

    out.print LEV[0],'<tr align="center">'
    out.print LEV[1],'<td></td>'
    out.print LEV[1],'<td> Platz </td>'
    out.print LEV[1],'<td> Schnitt </td>'
    out.print LEV[1],'<td> Buchholz. </td>' if $options.use_buchholz
    out.print LEV[1],'<td> Tor Diff. </td>'
    out.print LEV[1],'<td> Off. </td>'
    out.print LEV[1],'<td> Def.</td>'
    out.print LEV[1],'<td> Punkte </td>'
    out.print LEV[1],'<td> Tore </td>'
    out.print LEV[1],'<td> Sp. </td>'
    out.print LEV[1],'<td> S </td>'
    out.print LEV[1],'<td> U </td>'
    out.print LEV[1],'<td> N </td>'
    if @arr.first.tmp
      #out.print '<td style="border-style:none; min-width=5cm"></td>'
      out.print '<td style="border-style:none; color:#ffffff"> __ </td>'
      out.print LEV[1],'<td colspan="',@arr.first.tmp.size,'"> bisherige Platzierungen </td>'
    end
    out.print LEV[0],'</tr>'

    @arr.each_with_index{ | p, idx |
      next if p.dummy

      out.print LEV[0],'<tr align="center"'
      if ( idx % 2 == 0 )
        #out.print ' bgcolor="#eeeeee"'
        out.print 'style="background-color:#eeeeee"'
      end
      out.print '>'

      #out.print LEV[1],'<td>' , idx+1 , '</td>'
      out.print LEV[1],'<td>' ,  player_name(p) , '</td>'

      #sign = if @results.get_goal_diff(player_idx) > 0 then '+' else '' end
      out.print LEV[1],'<td>' , p.real_rank ,'</td>'
      if (col == "avg")
	out.print LEV[1],'<td style="color:#ee0000">'
      else
	out.print LEV[1],'<td>' 
      end
      out.print p.stat.get_avg_points,'</td>'
      out.print LEV[1],'<td>', p.stat.get_buchholz_value,'</td>' if $options.use_buchholz
      out.print LEV[1],'<td>', p.stat.get_avg_goal_diff,'</td>'
      if (col == "off")
	out.print LEV[1],'<td style="color:#ee0000">'
      else
	out.print LEV[1],'<td>' 
      end
      out.print p.stat.get_avg_goals,'</td>'
      if (col == "def")
	out.print LEV[1],'<td style="color:#ee0000">'
      else
	out.print LEV[1],'<td>' 
      end
      out.print p.stat.get_avg_goals_against,'</td>'
      out.print LEV[1],'<td>' , p.stat.get_num_points ,':',p.stat.get_num_points_against,'</td>'
      out.print LEV[1],'<td>' , p.stat.num_goals ,':',p.stat.num_goals_against,'</td>'
      out.print LEV[1],'<td>' , p.stat.num_games,'</td>'
      out.print LEV[1],'<td>' , p.stat.num_wins,'</td>'
      out.print LEV[1],'<td>' , p.stat.num_draws,'</td>'
      out.print LEV[1],'<td>' , p.stat.num_defeats,'</td>'
      if p.tmp
        out.print LEV[1]
        #out.print '<td style="background-color:#00ff00"> </td>'
        #out.print '<td style="background-color:#333333"> </td>'
        out.print '<td style="border-top-style:none; border-bottom-style:none;background-color:#ffffff"></td>'
        p.tmp.each{ |r|
          out.print '<td>', r, '</td>'
        }
      end
      out.print LEV[0],'</tr>'
    }
    out.print LEV[0],'</tbody>'
    out.print LEV[0],'</table>'
    html.print_document_end(out)
  end
  
  private

  def sort_ranking_using_stat
    @arr.sort!{ |p1,p2| 
      next 1 if p1.dummy
      next -1 if p2.dummy
      p2.stat <=> p1.stat 
    }

    last= @arr.first
    last.real_rank= 1
    @arr.each_with_index{ | p , i |
      p.rank= i
      if (p.stat <=> last.stat) != 0 
        p.real_rank= i + 1
      else
        p.real_rank= last.real_rank
      end
      last= p
    }
  end

end

##############################################################################

class SwissSystem
  def initialize(ranking, games)
    @ranking= ranking
    @gamesDB= games
    @matrix= Array.new
    @show_matrix= Array.new
    @ranking.size.times {
      @matrix << Array.new(@ranking.size, false)
      @show_matrix << Array.new(@ranking.size, nil)
    }
    @matrix.each_with_index { |row,idx|
      row[idx]= true
    }     
    @gamesDB.each{ |g|
      #debug
      if @matrix[ g.p1.rank ][ g.p2.rank ] 
        raise "game repetition"
      end
      @matrix[ g.p1.rank ][ g.p2.rank ]= true
      @matrix[ g.p2.rank ][ g.p1.rank ]= true
      @show_matrix[ g.p1.rank ][ g.p2.rank ]= "#{g.res1}:#{g.res2}"
      @show_matrix[ g.p2.rank ][ g.p1.rank ]= "#{g.res2}:#{g.res1}"
      #@show_matrix[ g.p1.rank ][ g.p2.rank ]= "#{g.p1.rank + g.p2.rank}"
      #@show_matrix[ g.p2.rank ][ g.p1.rank ]= "#{g.p1.rank + g.p2.rank}"
    }
    #show
  end
  
  def generate_games
    size= @matrix.size
    row_in_col= Array.new(size, -1)
    has_pair= Array.new(size, false)
    pairing_stack= []

    col_idx= 0
    row_idx= 0
    print "\ngenerate begin ..."
    i= 0
    loop do
      i+=1
      if row_idx == nil
        row_idx= first_free_row( has_pair )
        col_idx= row_idx
        #print " ", row_idx, "(", pairing_stack.size,")"
      end

      if row_idx == nil
        print " done"
        print "\nfound all pairings (#{i})"
        break
      end

      col_idx= next_free_col_in_row( row_idx, col_idx, has_pair )
      if col_idx != nil
        #found a pairing
        has_pair[ row_idx ]= true
        has_pair[ col_idx ]= true
        @matrix[row_idx][col_idx]= true
        @matrix[col_idx][row_idx]= true
        pairing_stack.push( [row_idx, col_idx] )
        row_idx= nil
        col_idx= -1
        next
      end
      #no pairing for this row
      row_idx, col_idx= pairing_stack.pop
      if row_idx == nil
        raise "\n no further games are possible"
      end
      has_pair[row_idx]= false
      has_pair[col_idx]= false
      @matrix[row_idx][col_idx]= false
      @matrix[col_idx][row_idx]= false
      #notice: col_idx is now set as the last entry of row_index
      #during the next iteration the following col_idx will be
      #inspected
    end

    #show the new games
    res= GamesDB.new
    res.round= @gamesDB.round+1

    pairing_stack.each { |row,col|
      #print "\ngame ", " : ", @ranking[row].rank, " vs ", @ranking[col].rank
      p1= @ranking[row]
      p2= @ranking[col]
      if p1.white_black_balance > p2.white_black_balance
        p1, p2= p2, p1
      end

      res.put( Game.new( p1, p2, nil, nil ) )
      #p1.white_black_balance += 1 unless p2.dummy
      #p2.white_black_balance -= 1 unless p1.dummy
    }

    res.make_schedule
    return res
  end


  def show
    print "\nSwissSystem"
    print "\nMatrix:"
    @matrix.each_with_index { |row,i|
      print "\n%3d" % i , "|"
      row.each_with_index { |e,j|
        print " X" if e
        print " ." if not e
      }
    }
  end

  def show_html( out  )
    col= nil
    html= HTML.new
    html.fontsize= $options.fontsize_crosstable
    html.print_document_begin(out)
    num= @gamesDB.round
    out.print "<h1> Kreuztabelle nach #{num} Runden </h1>"

    out.print '<table>'
    out.print '<tbody>'
    out.print '<tr><td colspan="2"/>'
    @ranking.size.times{|j|
      style= ' style="background-color:#f5f5f5"'
      style= '' if j % 2 != 0
      out.print "<td#{style}>",j+1,"</td>"
    }
    out.print '</tr>'
    @show_matrix.each_with_index { |row,i|
      out.print "\n<tr>"

      row.each_with_index { |e,j|
        if i == j
          style= ' style="background-color:#999999"'
        elsif i % 2 == 0 
          style= ' style="background-color:#dddddd"'
        elsif j % 2 == 0 
          style= ' style="background-color:#f5f5f5"'
        else
          style=''
        end

        style2= ' style="background-color:#dddddd; white-space: nowrap;"'
        if j == 0
          style2= ' style="white-space: nowrap;"' if i % 2 != 0
          out.print "\n<td#{style2}>", i+1, "</td>"
          out.print "\n<td#{style2}>", player_name(@ranking[i]), "</td>"
        end

        if i == j
          out.print "<td#{style}></td>" 
        elsif e
          out.print "<td#{style}>",e,"</td>"
        else
          out.print "<td#{style}/>"
        end
      }
      out.print "</tr>"
    }
    out.print '</tbody>'
    out.print '</table>'
    html.print_document_end(out)
  end

  private


  def next_free_col_in_row( row_idx, col_idx, has_pair)
    size= @matrix.size
    row= @matrix[row_idx]
    loop do
      col_idx += 1
      return nil if col_idx >= size 
      return col_idx if row[col_idx] == false and has_pair[col_idx] == false
    end
  end
  
  def first_free_row( has_pair)
    size= @matrix.size
    row_idx= 0
    loop do
      return nil if row_idx >= size 
      return row_idx if has_pair[row_idx] == false
      row_idx += 1
    end
  end
end

##############################################################################
##############################################################################
##############################################################################

if $0 == __FILE__
  #command line invocation
  options_fname= "options.yaml"
  if ARGV.size >= 1
    options_fname= ARGV[0] 
  elsif ARGV.size != 0
    print "usage: ruby ssys.rb [options_fname [num_iterations]]"
  end
  
  $options= YAML::load( File.open(options_fname) )

  steps= 1
  steps= ARGV[1].to_i if ARGV[1]

  unless File.exists?( $options.output_dir )
    print "\ngenerating directory: ", $options.output_dir
    Dir.mkdir( $options.output_dir )
  end

  #debug
  
(1..steps).each do |s|
  ignore_games_file= true if steps > 1 and s == 1

  playersDB_fname= $options.input_dir+"participants.yaml"
  dtkv_ranking_fname= $options.input_dir+"dtkv_ranking.yaml"
  dtkv_ranking_fname= $options.input_dir+"r_all.csv"

  gamesDB_fname= $options.output_dir+"games.txt"


  unless File.exists?(dtkv_ranking_fname)
    print <<EOS

Die Datei #{dtkv_ranking_fname} fehlt und muss von 

http://www.artbot.de/dtkv/download.html

heruntergeladen und im Verzeichnis 

#{$options.input_dir}

abgelegt werden!
EOS
    exit

  end
  dtkv_ranking= DTKVranking.new(dtkv_ranking_fname)
  playersDB= PlayersDB.new(playersDB_fname, dtkv_ranking)
  gamesDB= GamesDB.new(playersDB)
  ranking= Ranking.new(playersDB)
  out_ranking= 'ranking.html'

  if File.exists?(gamesDB_fname) and not ignore_games_file
    gamesDB.read(gamesDB_fname)
    ranking.compute_ranking_from_games(gamesDB)
    File.open($options.output_dir+out_ranking,'w'){|out|
      ranking.show_html( out, gamesDB.round )
    }
    ranking.show
    #gamesDB.debug
    File.open($options.output_dir+'ranking.yml','w') {|out|
      ranking.show_yaml( out )
    }
  else
    ranking.compute_ranking_from_dtkv_ranking
    #ranking.show
    File.open($options.output_dir+out_ranking,"w"){|out|
      ranking.show_html( out )
    }
    ranking.interlace_ranking #in first round 1 plays vs n/2 etc
    #ranking.show
  end

#ranking.show
#playersDB.show

  def round_str(round) 
    return "0"+round.to_s if round < 10
    return round.to_s
  end



  print "\n===================================="

  size= ranking.size
  print "\ngames for #{gamesDB.round+1}. round"
  swissSystem= SwissSystem.new( ranking, gamesDB )
  File.open($options.output_dir+"matrix.html","w"){|out|
    swissSystem.show_html(out)
  }
  #exit
  games= swissSystem.generate_games
  games.round= gamesDB.round+1
  games.generate_random_results if $options.random_results

  File.open($options.output_dir+"games_new.html","w"){|out|
    games.show_html(out) 
  }
  
  if gamesDB.round > 0
    #file= File.open("g_all#{gamesDB.round}.html","w")
    File.open($options.output_dir+"games_last.html","w") {|out|
      gamesDB.show_html(out, :last) 
    }
    File.open($options.output_dir+"games_all.html","w") {|out|
      gamesDB.show_html(out, :all) 
    }
    File.open($options.output_dir+"backup_games_%s.txt" % round_str(gamesDB.round),"w") {|out|
      gamesDB.show_html(out, :all) 
    }
  end

  gamesDB.concat(games)
  gamesDB.write(gamesDB_fname)
  gamesDB.write($options.output_dir+"backup_games_%s.txt" % round_str(gamesDB.round) )

end

end

